////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2001-2024 Alexey Kuryakin daqgroup@mail.ru under MIT license //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// This file is part of the CRW-DAQ project by DaqGroup - component CRWLIB.   //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Purpose:                                                                   //
// This unit provides wrap classes to embed Lazarus FPC compiler into your    //
// application and to create plugin DLLs, which can expand application        //
// features. Uses FpcupDeluxe compiler package.                               //
// See https://wiki.freepascal.org/fpcupdeluxe                                //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:                                                                   //
// 20241009 - Creation, uses _crw_dcc32.pas                                   //
////////////////////////////////////////////////////////////////////////////////

unit _crw_fpcup; // fpcupdeluxe wrapper

{$I _crw_sysdef}

{$IFDEF FPC}{$mode Delphi}{$ENDIF}{$H+}

{$WARN 5023 off : Unit "$1" not used in $2}

interface

uses
 _crw_sharemem, // NB: THIS UNIT MUST BE FIRST !!!
 {$IFDEF UNIX} cthreads, dl, {$ENDIF}
 {$IFDEF WINDOWS} windows, {$ENDIF}
 sysutils, classes, math,
 _crw_alloc, _crw_fpu, _crw_ef, _crw_str, _crw_fio, _crw_proc, _crw_gloss,
 _crw_sect, _crw_task, _crw_utf8, _crw_apputils, _crw_crwapi;

 ///////////////////////////////////////////////////////////////////////////////
 // TPluginWrapper can help you to load plugin DLL and run plugin function,   //
 // which exported by this DLL. Plugin function have signature:               //
 //  function(Api:TPluginApi):Integer; StdCall;                               //
 // Api class provides capabilities for main program <--> plugin interaction. //
 // Api class must be abstract virtual class, I mean, all members of this     //
 // class must be functions/procedures marked as "virtual; abstract;"         //
 // Main program create inherited Api class to overload abstract methods, but //
 // plugin knows only abstract Api. Plugin function executes in caller        //
 // context (address space, thread etc), but can use only methods provided by //
 // Api class.                                                                //
 // Publuc methods are:                                                       //
 //  DllInst    Return DLL instance <> 0, or 0 if DLL was not loaded          //
 //  DllPath    Return DLL file path                                          //
 //  PluginName Return plugin function name                                   //
 //  DllLoad    Load DLL, find plugin function address.                       //
 //             Return DLL instance <> 0 if Ok or 0 if any error.             //
 //  DllFree    Free DLL, if one was loaded.                                  //
 //  Execute    Load DLL, if need, and execute plugin function.               //
 //             Return Integer result of plugin function in lower 32 bits.    //
 //             Return Int64(-1) if any error.                                //
 // Example:                                                                  //
 //  Plugin DLL:                                                              //
 //   Library PluginLib;                                                      //
 //   uses _CrwApi;                                                           //
 //   function PluginExample(Api:TPluginApi):Integer; StdCall;                //
 //   begin                                                                   //
 //    with Api,SysApi do                                                     //
 //    if InitStdOut(Output) then begin                                       //
 //     writeln('Message from Plugin DLL!');                                  //
 //    end;                                                                   //
 //   end;                                                                    //
 //  Main program:                                                            //
 //   Api:=TCrwApiServer.Create;                                              //
 //   Plugin:=NewPluginWrapper('PluginLib.dll','PluginExample');              //
 //   Code:=Plugin.Execute(Api);                                              //
 //   if Code=-1 then writeln('Error') else writeln('Result=',Integer(Code)); //
 //   Kill(Plugin);                                                           //
 ///////////////////////////////////////////////////////////////////////////////
type
 EFpcupPluginFailure = class(EEchoException);
 TFpcupPluginWrapper = class(TMasterObject)
 private
  myDllInst : HMODULE;
  myDllPath : LongString;
  myPluginF : TPluginFun;
  myPluginN : LongString;
  function    GetDllInst:HMODULE;
  function    GetDllPath:LongString;
  function    GetPluginName:LongString;
 protected
  procedure   ErrorFound(E:Exception; const Note:LongString=''); override;
 public
  constructor Create(const aDllPath,aPluginName:LongString);
  destructor  Destroy; override;
  property    DllInst    : HMODULE     read GetDllInst;
  property    DllPath    : LongString  read GetDllPath;
  property    PluginName : LongString  read GetPluginName;
  function    DllLoad    : HMODULE;
  procedure   DllFree;
  function    Execute(Api:TPluginApi):Int64;
 end;

function  NewFpcupPluginWrapper(const aDllPath,aPluginName:LongString):TFpcupPluginWrapper;
procedure Kill(var TheObject:TFpcupPluginWrapper); overload;

 ///////////////////////////////////////////////////////////////////////////////
 // TFpcupWrapper provides access to Fpcup command line compiler to use       //
 // this compiler as embeded in your applications.                            //
 // BinDir                      Full path of ..\bin subdirectory              //
 // LibDir                      Full path of ..\lib subdirectory              //
 // ExeDir                      Full path of ..\exe subdirectory              //
 // TmpDir                      Full path of ..\tmp subdirectory              //
 // StdOut                      After compilation this text contains hooked   //
 //                             compiler messages.                            //
 // Hints                       After compilation this text contains hooked   //
 //                             compiler hints.                               //
 // Warnings                    After compilation this text contains hooked   //
 //                             compiler warnings.                            //
 // Errors                      After compilation this text contains hooked   //
 //                             compiler errors.                              //
 // Fatals                      After compilation this text contains hooked   //
 //                             compiler fatals.                              //
 // CanCompile(aProject)        Check compiler integrity & project existance  //
 // TryRemedy(aProject)         Try remedy compiler if one was damaged.       //
 //                             Use RemedyCmd command to remedy, for example  //
 //                             Resource\Remedy\FpcupBin.cmd Resource\Fpcup\  //
 // Compile(Proj,Dir,Ext)       Compile project file Proj. If Dir <> '', move //
 //                             compiled file Proj.Ext from ..\exe directoty  //
 //                             to Dir.                                       //
 // What compile method doing:                                                //
 // Before compile, set current directory to ..\bin and writes Config text to //
 // Fpcup.cfg. Compile always to ..\exe, send *.dcu to ..\tmp. Redirect Fpcup //
 // output to temporary file and read this one after compilation. Then parse  //
 // hooked Fpcup messages to Hints, Warnings, Errors, Fatals. For details use //
 // ParseFpcupMessage function. Then, if destination directory specified,     //
 // move compiled file to this one.                                           //
 ///////////////////////////////////////////////////////////////////////////////

type
 TFpcupWrapper=class;
 TFpcupStateCode=(scStandBy,scRunning,scSucceed,scFailed);
 TFpcupLineParser=procedure(F:TFpcupWrapper; const Line:LongString);
 EFpcupFailure = class(EEchoException);
 TFpcupWrapper = class(TMasterObject)
 private
  myLazBuildTid      : Integer;
  myLazBuildCmd      : LongString;
  myLazBuildLpr      : LongString;
  myLazBuildBuf      : LongString;
  myLazBuildLinking  : LongString;
  myLazBuildExitCode : Integer;
  myCompilerExitCode : Integer;
  myLineParser       : TFpcupLineParser;
  myStdOut           : TText;
  myHints            : TText;
  myWarnings         : TText;
  myErrors           : TText;
  myFatals           : TText;
  function    GetLazBuildTid:Integer;
  function    GetLazBuildCmd:LongString;
  function    GetLazBuildLpr:LongString;
  function    GetLazBuildLinking:LongString;
  function    GetLazBuildBusy:Boolean;
  function    GetLazBuildState:TFpcupStateCode;
  function    FreeLazBuildTid:Boolean;
  procedure   ClearLazBuildLists;
  function    GetStdOut:TText;
  function    GetHints:TText;
  function    GetWarnings:TText;
  function    GetErrors:TText;
  function    GetFatals:TText;
  function    GetLazBuildExitCode:Integer;
  function    GetCompilerExitCode:Integer;
 protected
  procedure   ErrorFound(E:Exception; const Note:LongString=''); override;
 public
  constructor Create(const aLazBuild:LongString='');
  destructor  Destroy; override;
  property    LazBuildTid      : Integer         read GetLazBuildTid;
  property    LazBuildCmd      : LongString      read GetLazBuildCmd;
  property    LazBuildLpr      : LongString      read GetLazBuildLpr;
  property    LazBuildLinking  : LongString      read GetLazBuildLinking;
  property    LazBuildBusy     : Boolean         read GetLazBuildBusy;
  property    LazBuildState    : TFpcupStateCode read GetLazBuildState;
  property    LazBuildExitCode : Integer         read GetLazBuildExitCode;
  property    CompilerExitCode : Integer         read GetCompilerExitCode;
  property    StdOut           : TText           read GetStdOut;
  property    Hints            : TText           read GetHints;
  property    Warnings         : TText           read GetWarnings;
  property    Errors           : TText           read GetErrors;
  property    Fatals           : TText           read GetFatals;
 public
  function    CanCompile(aProject:LongString=''):Boolean;
  function    Compile(const aProject:LongString; aParser:TFpcupLineParser=nil):Boolean;
  function    ReportLazBuild:LongString;
  procedure   ResetLazBuild;
 public
  procedure   StartPolling;
  procedure   StopPolling;
  procedure   DoPolling;
  procedure   DoReading;
  function    LazHelp(const SourceFile:LongString):Boolean;
  function    LazarusCmdLine:LongString;
  function    LazarusInstalled:Boolean;
 public
  class var StdIoPipeSizeKb:Integer;
  class function fwStateName(State:TFpcupStateCode):LongString;
 end;

function  NewFpcupWrapper(const aFpcupDir:LongString=''):TFpcupWrapper;
procedure Kill(var TheObject:TFpcupWrapper); overload;

procedure DefaultFpcupLineParser(F:TFpcupWrapper; const Line:LongString);

 ///////////////////////////////////////////////////////////////////////////////
 // The only one instance of Fpcup.                                           //
 ///////////////////////////////////////////////////////////////////////////////
function  NewFpcupWrapperFromIniFile:TFpcupWrapper;

 ///////////////////////////////////////////////////////////////////////////////
 // Use Fpcup function to compile your projects easy.                         //
 // Fpcup automatically creates TFpcupWrapper object if one does not exists,  //
 // so you should not declare any variables. Created object will be destroyed //
 // on program exit. You also may Fpcup.Free if you don't need Fpcup any more.//
 // Example:                                                                  //
 //  Fpcup.Compile('plugin1.lpr');  // Create FpcupWrapper and compile        //
 //  Fpcup.Free;                    // Destroy wrapper                        //
 //  ...                                                                      //
 //  Fpcup.Compile('plugin2.lpr');  // Create wrapper again and compile       //
 ///////////////////////////////////////////////////////////////////////////////
function  Fpcup:TFpcupWrapper;

 ///////////////////////////////////////////////////////////////////////////////
 // Use ParseFpcupMessage function to parse Fpcup output hints, warnings,     //
 // errors and fatals. For example:                                           //
 //  Fpcup.Compile('proj.lpr');                                               //
 //  for i:=0 to Fpcup.Hints.Count-1 do                                       //
 //  with ParseFpcupMessage(Fpcup.Hints[i]) do                                //
 //  writeln('Hint File:',FileName,' Line:',LineNum,' Token:',Token);         //
 ///////////////////////////////////////////////////////////////////////////////
type
 TParseFpcupStatus  = (cmsNone,cmsHint,cmsWarning,cmsError,cmsFatal,cmsLinking,cmsCompiled,cmsExitCode);
 TParseFpcupMessage = packed record
  Status   : TParseFpcupStatus;
  LineNum  : Integer;
  ColNum   : Integer;
  Compiled : Integer;
  ExitCode : Integer;
  FileName : PureString;
  Token    : PureString;
 end;

function ParseFpcupMessage(const aMessage:LongString):TParseFpcupMessage;

procedure Timer_FpcupPolling;

implementation

uses LazUtf8;

{$IFDEF SKIP_DRAFT}
Complete this unit
{$ENDIF ~SKIP_DRAFT}

 ////////////////////////////////////////
 // TFpcupPluginWrapper implementation //
 ////////////////////////////////////////
constructor TFpcupPluginWrapper.Create(const aDllPath,aPluginName:LongString);
begin
 inherited Create;
 myDllInst:=0;
 myDllPath:=UnifyFileAlias(AdaptDllFileName(DefaultExtension(Trim(aDllPath),'.dll')));
 myPluginF:=nil;
 myPluginN:=Trim(aPluginName);
 Exceptions:=false;
 ErrorReportProc:=DefaultObjectErrorReportProc;
end;

destructor TFpcupPluginWrapper.Destroy;
begin
 DllFree;
 myDllPath:='';
 myPluginN:='';
 inherited Destroy;
end;

function TFpcupPluginWrapper.GetDllInst:HMODULE;
begin
 if Assigned(Self) then Result:=myDllInst else Result:=0;
end;

function TFpcupPluginWrapper.GetDllPath:LongString;
begin
 if Assigned(Self) then Result:=myDllPath else Result:='';
end;

function TFpcupPluginWrapper.GetPluginName:LongString;
begin
 if Assigned(Self) then Result:=myPluginN else Result:='';
end;

procedure TFpcupPluginWrapper.ErrorFound(E:Exception; const Note:LongString);
begin
 if Exceptions then begin
  if (E is Exception)
  then RAISE EFpcupPluginFailure.Create(E.Message)
  else RAISE EFpcupPluginFailure.Create(Note);
 end else ErrorReport(E,Note);
end;

function TFpcupPluginWrapper.DllLoad:HMODULE;
begin
 Result:=0;
 if Assigned(Self) then
 try
  myPluginF:=nil;
  if (DllInst=0) then begin
   if FileExists(DllPath)
   then myDllInst:=SafeLoadLibrary(DllPath)
   else RAISE EFpcupPluginFailure.Create(Format('File "%s" not found.',[DllPath]));
  end;
  if (DllInst=0)
  then RAISE EFpcupPluginFailure.Create(Format('Error LoadLibrary("%s"): "%s".',
                                     [DllPath,SysErrorMessage(GetLastOsError)]));
  myPluginF:=TPluginFun(System.GetProcAddress(DllInst,PluginName));
  if Assigned(myPluginF)
  then Result:=DllInst
  else RAISE EFpcupPluginFailure.Create(Format('Function "%s" not found  in "%s".',
                                     [PluginName,DllPath]));
 except
  on E:Exception do begin
   DllFree;
   ErrorFound(E,'DllLoad');
  end;
 end;
end;

procedure TFpcupPluginWrapper.DllFree;
begin
 if Assigned(Self) then
 try
  try
   if DllInst <> 0 then
   if not FreeLibrary(DllInst)
   then RAISE EFpcupPluginFailure.Create(Format('Error FreeLibrary("%s"): "%s".',
                                      [DllPath,SysErrorMessage(GetLastOsError)]));
  finally
   myPluginF:=nil;
   myDllInst:=0;
  end;
 except
  on E:Exception do ErrorFound(E,'DllFree');
 end;
end;

function TFpcupPluginWrapper.Execute(Api:TPluginApi):Int64;
begin
 Result:=-1;
 if Assigned(Api) then
 if Assigned(Self) then
 try
  try
   if not Assigned(myPluginF) then DllLoad;
   if Assigned(myPluginF) then begin
    Api.ServerActionsBeforeExecution;
    try
     Result:=myPluginF(Api);
    finally
     Api.ServerActionsAfterExecution;
    end;
   end;
  finally
   FpuSetExceptions(false);
  end;
 except
  on E:Exception do ErrorFound(E,'Execute');
 end;
end;

function NewFpcupPluginWrapper(const aDllPath,aPluginName:LongString):TFpcupPluginWrapper;
begin
 Result:=nil;
 try
  Result:=TFpcupPluginWrapper.Create(aDllPath,aPluginName);
 except
  on E:Exception do BugReport(E,nil,'NewFpcupPluginWrapper');
 end;
end;

procedure Kill(var TheObject:TFpcupPluginWrapper); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end;
end;

 //////////////////////////////////
 // TFpcupWrapper implementation //
 //////////////////////////////////

constructor TFpcupWrapper.Create(const aLazBuild:LongString='');
begin
 inherited Create;
 myLazBuildTid:=0;
 myLineParser:=DefaultFpcupLineParser;
 myLazBuildCmd:=UnifyFileAlias(aLazBuild);
 myLazBuildExitCode:=0; myCompilerExitCode:=0;
 myLazBuildLpr:=''; myLazBuildBuf:=''; myLazBuildLinking:='';
 myStdOut:=NewText;
 myStdOut.Master:=@myStdOut;
 myHints:=NewText;
 myHints.Master:=@myHints;
 myWarnings:=NewText;
 myWarnings.Master:=@myWarnings;
 myErrors:=NewText;
 myErrors.Master:=@myErrors;
 myFatals:=NewText;
 myFatals.Master:=@myFatals;
 Exceptions:=false;
 ErrorReportProc:=DefaultObjectErrorReportProc;
end;

destructor  TFpcupWrapper.Destroy;
begin
 FreeLazBuildTid;
 myLazBuildCmd:='';
 myLazBuildLpr:='';
 myLazBuildBuf:='';
 myLazBuildLinking:='';
 Kill(myStdOut);
 Kill(myHints);
 Kill(myWarnings);
 Kill(myErrors);
 Kill(myFatals);
 inherited Destroy;
end;

function TFpcupWrapper.FreeLazBuildTid:Boolean;
begin
 Result:=false;
 if Assigned(Self) then begin
  if (myLazBuildTid<>0) then begin
   Result:=task_free(myLazBuildTid);
   myLazBuildTid:=0;
  end;
  myLazBuildBuf:='';
 end;
end;

procedure TFpcupWrapper.ClearLazBuildLists;
begin
 if Assigned(Self) then begin
  StdOut.Count:=0;
  Hints.Count:=0;
  Warnings.Count:=0;
  Errors.Count:=0;
  Fatals.Count:=0;
 end;
end;

procedure TFpcupWrapper.ResetLazBuild;
begin
 if Assigned(Self) then begin
  StopPolling;
  FreeLazBuildTid;
  ClearLazBuildLists;
  myLazBuildExitCode:=0;
  myCompilerExitCode:=0;
  myLazBuildLinking:='';
  myLazBuildLpr:='';
  myLazBuildBuf:='';
 end;
end;

function TFpcupWrapper.GetLazBuildTid:Integer;
begin
 if Assigned(Self)
 then Result:=myLazBuildTid
 else Result:=0;
end;

function TFpcupWrapper.GetLazBuildCmd:LongString;
begin
 if Assigned(Self)
 then Result:=myLazBuildCmd
 else Result:='';
end;

function TFpcupWrapper.GetLazBuildLpr:LongString;
begin
 if Assigned(Self)
 then Result:=myLazBuildLpr
 else Result:='';
end;

function TFpcupWrapper.GetLazBuildLinking:LongString;
begin
 if Assigned(Self)
 then Result:=myLazBuildLinking
 else Result:='';
end;

function TFpcupWrapper.GetLazBuildExitCode:Integer;
begin
 if Assigned(Self)
 then Result:=myLazBuildExitCode
 else Result:=0;
end;

function TFpcupWrapper.GetCompilerExitCode:Integer;
begin
 if Assigned(Self)
 then Result:=myCompilerExitCode
 else Result:=0;
end;

function TFpcupWrapper.GetLazBuildBusy:Boolean;
begin
 if Assigned(Self)
 then Result:=(myLazBuildTid<>0) and task_wait(myLazBuildTid,0)
 else Result:=false;
end;

function TFpcupWrapper.GetStdOut:TText;
begin
 if Assigned(Self) then Result:=myStdOut else Result:=nil;
end;

function TFpcupWrapper.GetHints:TText;
begin
 if Assigned(Self) then Result:=myHints else Result:=nil;
end;

function TFpcupWrapper.GetWarnings:TText;
begin
 if Assigned(Self) then Result:=myWarnings else Result:=nil;
end;

function TFpcupWrapper.GetErrors:TText;
begin
 if Assigned(Self) then Result:=myErrors else Result:=nil;
end;

function TFpcupWrapper.GetFatals:TText;
begin
 if Assigned(Self) then Result:=myFatals else Result:=nil;
end;

procedure TFpcupWrapper.ErrorFound(E:Exception; const Note:LongString);
begin
 if Exceptions then begin
  if E is Exception
  then RAISE EFpcupFailure.Create(E.Message)
  else RAISE EFpcupFailure.Create(Note);
 end else ErrorReport(E,Note);
end;

function TFpcupWrapper.GetLazBuildState:TFpcupStateCode;
var tid:Integer;
begin
 Result:=scStandby;
 if Assigned(Self) then begin
  tid:=myLazBuildTid;
  if (tid=0) and (myLazBuildLpr='') then Exit;
  if (task_pid(tid)=0) then Exit(scStandby);
  if task_wait(myLazBuildTid,0) then Exit(scRunning);
  if (task_rxcount(tid)>0) then DoReading;
  if (myLazBuildExitCode<>0) then Exit(scFailed);
  if (myCompilerExitCode<>0) then Exit(scFailed);
  if (Errors.Count+Fatals.Count>0) then Exit(scFailed);
  Result:=scSucceed;
 end;
end;

function TFpcupWrapper.CanCompile(aProject:LongString=''):Boolean;
begin
 Result:=false;
 if Assigned(Self) then
 try
  // проверить сценарий компилятора
  if IsEmptyStr(LazBuildCmd) then RAISE EFpcupFailure.Create(SectFpcupDeluxe+' LazBuild not specified.');
  if not FileExists(LazBuildCmd) then RAISE EFpcupFailure.Create(LazBuildCmd+' is not found.');
  if not FileIsExecutable(LazBuildCmd) then RAISE EFpcupFailure.Create(LazBuildCmd+' is not executable.');
  // проверить файл проекта
  if not IsEmptyStr(aProject) then begin
   aProject:=UnifyFileAlias(aProject);
   if not FileExists(aProject)
   then RAISE EFpcupFailure.Create(Format('File "%s" not found.',[aProject]));
   if not SameText(ExtractFileExt(aProject),'.lpr')
   then RAISE EFpcupFailure.Create(Format('Invalid extension in "%s".',[aProject]));
  end;
  Result:=true;
 except
  on E:Exception do ErrorFound(E,'CanCompile');
 end;
end;

function TFpcupWrapper.Compile(const aProject:LongString; aParser:TFpcupLineParser=nil):Boolean;
var CmdLine:LongString; PipeSizeKb:Integer;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if (LazBuildState in [scRunning])
  then RAISE EFpcupFailure.Create(RusEng('Компилятор уже занят.','Compiler is already busy.'));
  // Check compiler & project exists
  if not CanCompile(aProject)
  then RAISE EFpcupFailure.Create(Format('Could not compile "%s".',[aProject]));
  if not Assigned(aParser) then aParser:=DefaultFpcupLineParser;
  myLineParser:=aParser;
  // Prepare variables
  FreeLazBuildTid; ResetLazBuild;
  myLazBuildLpr:=UnifyFileAlias(aProject);
  CmdLine:=Format('%s %s',[LazBuildCmd,myLazBuildLpr]);
  // Запуск компилятора
  myLazBuildTid:=task_init(CmdLine);
  PipeSizeKb:=EnsureRange(StdIoPipeSizeKb,16,KiloByte*16);
  task_ctrl(myLazBuildTid,'StdInpPipeSize='+IntToStr(KiloByte*PipeSizeKb));
  task_ctrl(myLazBuildTid,'StdOutPipeSize='+IntToStr(KiloByte*PipeSizeKb));
  task_ctrl(myLazBuildTid,'Display='+IntToStr(SW_SHOWMINNOACTIVE));
  Result:=task_run(myLazBuildTid);
  if Result then begin
   StartPolling;
  end else begin
   StopPolling; FreeLazBuildTid; ResetLazBuild;
   RAISE EFpcupFailure.Create('Could not start compiler.');
  end;
 except
  on E:Exception do ErrorFound(E,'Compile');
 end;
end;

class function TFpcupWrapper.fwStateName(State:TFpcupStateCode):LongString;
begin
 case State of
  scStandby   : Result:=RusEng('Ожидание','Standby');
  scRunning   : Result:=RusEng('Работает','Running');
  scSucceed   : Result:=RusEng('Успешно','Succeed');
  scFailed    : Result:=RusEng('Ошибка','Failed');
  else          Result:=RusEng('Сбой','Fatal');
 end;
end;

function TFpcupWrapper.ReportLazBuild:LongString;
const sep=',  '; var i:Integer;
begin
 Result:='';
 if Assigned(Self) then begin
  Result:=RusEng('Отчёт компиляции:','Compiler report:')+EOL
         +RusEng('#################','################')+EOL
         +RusEng(' Проект: ',' Source: ')+LazBuildLpr+EOL;
  if (LazBuildLinking<>'') then Result:=Result
         +RusEng(' Сборка: ',' Linked: ')+LazBuildLinking+EOL;
  Result:=Result
         +RusEng(' Статус: ',' Status: ')+utf8_uppercase(fwStateName(LazBuildState))+sep
         +RusEng(' КодКомпиляции: ',' CompilerExitCode: ')+IntToStr(Fpcup.CompilerExitCode)+sep
         +RusEng(' КодСценария: ',' ScriptExitCode: ')+IntToStr(Fpcup.LazBuildExitCode)+sep
         +RusEng(' Ошибок: ',' Errors: ')+IntToStr(Fpcup.Errors.Count)+sep
         +RusEng(' Сбоев: ',' Fatals: ')+IntToStr(Fpcup.Errors.Count)+sep
         +RusEng(' Предупреждений: ',' Warnings: ')+IntToStr(Fpcup.Warnings.Count)+sep
         +RusEng(' Советов: ',' Hints: ')+IntToStr(Fpcup.Hints.Count)+EOL;
  if (Errors.Count+Fatals.Count>0) then begin
   Result:=Result+RusEng(' Список ошибок и сбоев:',' List of Errors and Fatals:')+EOL;
   Result:=Result+RusEng(' ----------------------',' --------------------------')+EOL;
   for i:=0 to Errors.Count-1 do Result:=Result+'  '+Errors[i]+EOL;
   for i:=0 to Fatals.Count-1 do Result:=Result+'  '+Fatals[i]+EOL;
  end;
 end;
end;

function TFpcupWrapper.LazHelp(const SourceFile:LongString):Boolean;
var Path,exe,cmd:LongString;
begin
 Result:=false;
 try
  Path:='';
  if SysGlossary.ReadIniPath(SysIniFile,SectFpcupDeluxe,'LazHelp',HomeDir,Path) then begin
   exe:=UnifyFileAlias(AdaptExeFileName(Path));
   cmd:=AnsiQuotedIfNeed(exe)+' '+AnsiQuotedIfNeed(SourceFile);
   if FileExists(exe)
   then Result:=(SendToMainConsole('@silent @run -sw7 '+cmd+EOL)>0)
   else Echo(RusEng('Не найден: ','Not found: ')+exe);
  end;
 except
  on E:Exception do BugReport(E,Self,'LazHelp');
 end;
end;

function TFpcupWrapper.LazarusCmdLine:LongString;
var Buff,Lnk,Exec:LongString;
begin
 Result:='';
 try
  Lnk:='';
  if SysGlossary.ReadIniPath(SysIniFile,SectFpcupDeluxe,'FpcupLnk',HomeDir,Lnk) then begin
   Lnk:=UnifyFileAlias(AdaptLnkFileName(Lnk));
   Buff:=ReadShellLinkAsText(Lnk);
   Exec:=CookieScan(Buff,'Exec');
   Result:=Trim(Exec);
  end;
 except
  on E:Exception do BugReport(E,Self,'LazarusCmdLine');
 end;
end;

function TFpcupWrapper.LazarusInstalled:Boolean;
var exe:LongString;
begin
 Result:=False;
 exe:=ExtractPhrase(1,LazarusCmdLine,JustSpaces);
 if (exe<>'') and FileExists(exe) and FileIsExecutable(exe)
 then Result:=True;
end;

procedure TFpcupWrapper.StartPolling;
begin
 if Assigned(Self) then begin
  Tick55Actions.Add(Timer_FpcupPolling);
 end;
end;

procedure TFpcupWrapper.StopPolling;
begin
 if Assigned(Self) then begin
  if (myLazBuildTid<>0) then DoReading;
  Tick55Actions.Remove(Timer_FpcupPolling);
 end;
end;

function Task_Readln(tid:Integer; var Line,Buff:LongString):Boolean;
const MaxLeng=KiloByte*16;
var p:Integer;
begin
 Line:='';
 Task_Readln:=False;
 if (Task_Pid(tid)<>0) then begin
  if (Length(Buff)<MaxLeng) and (Task_RxCount(tid)>0)
  then Buff:=Buff+Task_Recv(tid,MaxLeng-Length(Buff));
  p:=PosEol(Buff,1,0);
  if (p>0) then begin
   Task_Readln:=True;
   if (p>1) then Line:=Copy(Buff,1,p-1);
   Buff:=Copy(Buff,PosEol(Buff,p,1),MaxInt);
  end else begin
   if (Length(Buff)>=MaxLeng) then begin
    Echo('Received line is too long!');
    Buff:='';
   end;
  end;
 end;
end;

function OnPollingLines(n:SizeInt; Line:LongString; Custom:Pointer):Boolean;
var Fpcup:TFpcupWrapper;
begin
 Result:=true;
 if Assigned(Custom) then begin
  Fpcup:=TFpcupWrapper(Custom);
  if Assigned(Fpcup.myLineParser) then begin
   Fpcup.myLineParser(Fpcup,Line);
  end else begin
   Fpcup.StdOut.AddLn(Line);
   Echo(Line);
  end;
 end;
end;

procedure TFpcupWrapper.DoReading;
var tid,len,n:Integer; line:LongString;
begin
 if Assigned(Self) then
 try
  tid:=myLazBuildTid;
  if (tid=0) then Exit;
  len:=task_rxcount(tid);
  if (len>0) then begin
   line:=''; n:=0;
   while task_readln(tid,Line,myLazBuildBuf) do begin
    OnPollingLines(n,line,Self);
    Inc(n);
   end;
  end;
  myLazBuildExitCode:=task_result(tid);
 except
  on E:Exception do BugReport(E,nil,'DoReading');
 end;
end;

procedure TFpcupWrapper.DoPolling;
var tid,pid:Integer;
begin
 if Assigned(Self) then begin
  tid:=myLazBuildTid; if (tid=0) then Exit;
  pid:=task_pid(tid); if (pid<>0) then DoReading else Exit;
  if task_wait(tid,0) then Exit else StopPolling;
 end;
end;

procedure Timer_FpcupPolling;
begin
 Fpcup.DoPolling;
end;

function CheckStatus(const Line,Key:LongString; s:TParseFpcupStatus;
                      var R:TParseFpcupMessage;  var p:Integer):Boolean;
begin
 Result:=false;
 if (R.Status=cmsNone) then begin
  p:=Pos(Key,Line);
  if p=0 then Exit;
  Result:=true;
  R.Status:=s;
 end;
end;

function ParseFpcupMessage(const aMessage:LongString):TParseFpcupMessage;
var p,i:Integer; lincol:LongString;
begin
 Result:=Default(TParseFpcupMessage); p:=0;
 if CheckStatus(aMessage, ' Fatal: ',           cmsFatal,    Result, p)
 or CheckStatus(aMessage, ' Error: ',           cmsError,    Result, p)
 or CheckStatus(aMessage, ' Warning: ',         cmsWarning,  Result, p)
 or CheckStatus(aMessage, ' Hint: ',            cmsHint,     Result, p)
 or CheckStatus(aMessage, ' Linking ',          cmsLinking,  Result, p)
 or CheckStatus(aMessage, ' lines compiled',    cmsCompiled, Result, p)
 or CheckStatus(aMessage, 'CompilerExitCode: ', cmsExitCode, Result, p)
 or CheckStatus(aMessage, ' stopped with exit code ', cmsExitCode, Result, p)
 then with Result do
 try
  case Status of
   cmsFatal,cmsError,cmsWarning,cmsHint: begin
    i:=p-1;
    while (i>1) and (StrFetch(aMessage,i)<>'(') do Dec(i);
    if (i>1) then begin
     FileName:=Copy(aMessage,1,i-1);
     LinCol:=Copy(aMessage,i+1,p-i-2);
     LineNum:=StrToIntDef(ExtractWord(1,LinCol,ScanSpaces),0);
     ColNum:=StrToIntDef(ExtractWord(2,LinCol,ScanSpaces),0);
    end;
    Token:=Trim(Copy(aMessage,p,Length(aMessage)-p+1));
   end;
   cmsLinking: begin
    FileName:=Trim(SkipWords(2,aMessage,ScanSpaces));
    Token:=Trim(Copy(aMessage,p,Length(aMessage)-p+1));
   end;
   cmsCompiled: begin
    Token:=Trim(SkipWords(1,aMessage,ScanSpaces));
    Compiled:=StrToIntDef(ExtractWord(1,Token,ScanSpaces),0);
   end;
   cmsExitCode: begin
    Token:=Trim(Copy(aMessage,p,Length(aMessage)-p+1));
    ExitCode:=StrToIntDef(ExtractWord(2,Token,ScanSpaces),-1);
    if (ExitCode=-1) then
    ExitCode:=StrToIntDef(ExtractWord(5,Token,ScanSpaces),-1);
   end;
  end;
 except
  on E:Exception do begin
   Status:=cmsFatal;
   FileName:=''; LineNum:=0; ColNum:=0;
   Token:='Fpcup output parser error.';
  end;
 end;
end;

function NewFpcupWrapper(const aFpcupDir:LongString=''):TFpcupWrapper;
begin
 Result:=nil;
 try
  Result:=TFpcupWrapper.Create(aFpcupDir);
 except
  on E:Exception do BugReport(E,nil,'NewFpcupWrapper');
 end;
end;

procedure Kill(var TheObject:TFpcupWrapper); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end; 
end;

procedure DefaultFpcupLineParser(F:TFpcupWrapper; const Line:LongString);
var Parse:TParseFpcupMessage;
begin
 Echo(Line);
 if (Line<>'') then
 if Assigned(F) then begin
  F.StdOut.AddLn(Line);
  Parse:=ParseFpcupMessage(Line);
  case Parse.Status of
   cmsHint     : F.Hints.AddLn(Line);
   cmsWarning  : F.Warnings.AddLn(Line);
   cmsError    : F.Errors.AddLn(Line);
   cmsFatal    : F.Fatals.Addln(Line);
   cmsLinking  : F.myLazBuildLinking:=Parse.FileName;
   cmsExitCode : F.myCompilerExitCode:=Parse.ExitCode;
  end;
 end;
end;

function DefaultLazBuild:LongString;
begin
 Result:=AddPathDelim(HomeDir)+AdaptExeFileName('resource\crwlazbuild.cmd');
 Result:=UnifyFileAlias(Result);
end;

function NewFpcupWrapperFromIniFile:TFpcupWrapper;
var s:LongString;
begin
 s:='';
 if ReadIniFilePath(SysIniFile,SectFpcupDeluxe,'LazBuild',HomeDir,s)
 then s:=UnifyFileAlias(AdaptExeFileName(s)) else s:=DefaultLazBuild;
 Result:=NewFpcupWrapper(s);
end;

const
 TheFpcup : TFpcupWrapper = nil;

function  Fpcup:TFpcupWrapper;
begin
 if not Assigned(TheFpcup) then begin
  TheFpcup:=NewFpcupWrapperFromIniFile;
  TheFpcup.Master:=@TheFpcup;
 end;
 Result:=TheFpcup;
end;

///////////////////////////////////////
// Unit initialization and finalization
///////////////////////////////////////

procedure Init_crw_fpcup;
begin
 TFpcupWrapper.StdIoPipeSizeKb:=64;
end;

procedure Free_crw_fpcup;
begin
 Kill(TheFpcup);
end;

initialization

 Init_crw_fpcup;

finalization

 Free_crw_fpcup;

end.

//////////////
// END OF FILE
//////////////

